/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada. 
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.util;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Vector;

import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JToggleButton;
import javax.swing.ListCellRenderer;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * The {@link JButtonList} class extends {@link JList} to allow {@link AbstractButton}
 * objects instead of the default {@link JLabel} objects.  
 * This class was originally designed to only use {@link JCheckBox} objects, but has now
 * been extended to support any kind of {@link AbstractButton} object.
 * <br>
 * If the model contains {@link AbstractButton} objects, then those are rendered.  Otherwise, an AbstractButton will 
 * be created for each object.  You can choose which type of button to use: {@link JCheckBox}, {@link JRadioButton}, 
 * {@link JToggleButton}, {@link JCheckBoxMenuItem}, etc using the {@link JButtonList#setButtonClass(Class)} method.
 * <br>
 * If you want to allow only one button to be selected at a time, call the {@link JButtonList#setRadioButtonMode(boolean)}. 
 * 
 * @author Chris Callendar
 */
public class JButtonList extends JList implements MouseListener, KeyListener, ListDataListener, ListSelectionListener, ItemListener {

	protected static final Border BORDER_NONE = new EmptyBorder(1, 1, 1, 1);
	
	private boolean singleClickSelectionChange;
	private boolean selectionChanging;
	private ArrayList/*<AbstractButton>*/ buttons = new ArrayList/*<AbstractButton>*/(0);
	private ExtendedButtonGroup group;
	private boolean radioButtonMode;
	private Class buttonClass;
	
	private ArrayList/*<ItemListener>*/ listeners = new ArrayList/*<ItemListener>*/();
	
	/**
	 * Initializes this object.  Adds mouse and key listeners
	 * and sets single selection mode and a CheckBoxCellRenderer.
	 */
	public JButtonList() {
		super();
		init();
	}

	/**
	 * @param dataModel
	 */
	public JButtonList(ListModel dataModel) {
		super(dataModel);
		init();
	}

	/**
	 * @param listData
	 */
	public JButtonList(Object[] listData) {
		super(listData);
		init();
	}

	/**
	 * @param listData
	 */
	public JButtonList(Vector/*<?>*/ listData) {
		super(listData);
		init();
	}
	
	private void init() {
		this.buttonClass = JCheckBox.class;
		this.group = new ExtendedButtonGroup();
		this.radioButtonMode = false;
		this.singleClickSelectionChange = true;
		this.selectionChanging = false;
		reload(false);
		getModel().addListDataListener(this);
		setCellRenderer(new CheckBoxCellRenderer());
		setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		addMouseListener(this);
		addKeyListener(this);
		addListSelectionListener(this);
	}
	
	///////////////////////////////////////////////
	// PUBLIC METHODS
	///////////////////////////////////////////////
	
	public void addItemListener(ItemListener il) {
		if (!listeners.contains(il)) {
			listeners.add(il);
		}
	}
	
	public void removeItemListener(ItemListener il) {
		listeners.remove(il);
	}
	
	public void addElement(Object element, boolean checked) {
		if (getModel() instanceof DefaultListModel) {
			DefaultListModel model = (DefaultListModel) getModel();
			model.addElement(element);
			AbstractButton button = getButton(model.getSize() - 1);
			if (button != null) {
				button.setSelected(checked);
				repaint();
			} else {
				System.err.println("Couldn't find the added button?");
			}
		}
	}
	
	public void setModel(ListModel model) {
		if (getModel() != null) {
			getModel().removeListDataListener(this);
		}
		super.setModel(model);
		if (getModel() != null) {
			getModel().addListDataListener(this);
		}
		reload(true);
	}
	
	/**
	 * Sets the button class to use.
	 * Must be a subclass of {@link AbstractButton}. 
	 * @param abstractButton
	 * @throws IllegalArgumentException if the class isn't a subclass of {@link AbstractButton}
	 */
	public void setButtonClass(Class/*<? extends AbstractButton>*/ abstractButton) throws IllegalArgumentException {
		if (abstractButton == null) {
			abstractButton = JCheckBox.class;
		}
		if (!abstractButton.equals(buttonClass)) {
			try {
				AbstractButton button = (AbstractButton) abstractButton.newInstance();
				button.setSelected(false);	// remove the warning
			} catch (Exception e) {
				throw new IllegalArgumentException("Invalid button class.  Must be a subclass of AbstractButton.", e);
			}
			this.buttonClass = abstractButton;
			
			// save the checked indeces
			int[] checked = getCheckedIndeces();
			reload(isVisible());
			// restore the checked indeces
			setCheckedIndices(checked);
		}
	}

	public Class getButtonClass() {
		return buttonClass;
	}
	
	/**
	 * If true is passed in then only one of the checkboxes/buttons can be 
	 * checked/selected at one time.
	 * @param onlyOneChecked
	 */
	public void setRadioButtonMode(boolean onlyOneChecked) {
		if (onlyOneChecked != radioButtonMode) {
			this.radioButtonMode = onlyOneChecked;
			// save the checked indeces
			int[] checked = getCheckedIndeces();
			reload(isVisible());
			// restore the checked indeces
			setCheckedIndices(checked);
		}
	}
	
	public boolean isRadioButtonMode() {
		return radioButtonMode;
	}
	
	/**
	 * If true is passed in then any time the mouse is clicked on an item 
	 * (the checkbox or the text) its checked value will be toggled.
	 * If false is passed in then the item's selected value will only toggle
	 * when either the checkbox is clicked or when the item is click when it is selected in the list. 
	 * @param singleClickSelectionChange
	 */
	public void setSingleClickSelectionChange(boolean singleClickSelectionChange) {
		this.singleClickSelectionChange = singleClickSelectionChange;
	}
	
	public boolean isSingleClickSelectionChange() {
		return singleClickSelectionChange;
	}
	
	/**
	 * @return an array of the objects in the list that are checked (selected).
	 */
	public Object[] getCheckedItems() {
		final Collection items = getCheckedItemsCollection();
		return items.toArray(new Object[items.size()]);
	}
	
	/**
	 * @return a {@link Collection} of the objects in the list that are checked (selected).
	 */
	public Collection/*<Object>*/ getCheckedItemsCollection() {
		ArrayList/*<Object>*/ checkedItems = new ArrayList/*<Object>*/();
		for (int i = 0; i < getModel().getSize(); i++) {
			Object obj = getModel().getElementAt(i);
			AbstractButton button = (AbstractButton) buttons.get(i);
			if (button.isSelected()) {
				checkedItems.add(obj);
			}
		}
		return checkedItems;
	}
	
	/**
	 * @return an array of the indeces in the list that are checked (selected).
	 */
	public int[] getCheckedIndeces() {
		ArrayList/*<Integer>*/ checkedIndeces = new ArrayList/*<Integer>*/();
		for (int i = 0; i < getModel().getSize(); i++) {
			AbstractButton button = (AbstractButton) buttons.get(i);
			if (button.isSelected()) {
				checkedIndeces.add(new Integer(i));
			}
		}
		int[] indeces = new int[checkedIndeces.size()];
		for (int i = 0; i < indeces.length; i++) {
			indeces[i] = ((Integer)checkedIndeces.get(i)).intValue();
		}
		return indeces;
	}
	
	/**
	 * Checks the buttons at the given indeces. 
	 * If null or an empty array is passed in then all the buttons are unchecked.
	 * If an invalid index is passed in nothing happens.
	 * @param checkedIndeces
	 */
	public void setCheckedIndices(int[] checkedIndeces) {
		if ((checkedIndeces != null) && (checkedIndeces.length > 0)) {
			for (int i = 0; i < checkedIndeces.length; i++) {
				int index = checkedIndeces[i];
				if ((index >= 0) && (index < buttons.size())) {
					((AbstractButton) buttons.get(index)).setSelected(true);
				}
			}
			
		} else {
			checkNone();
		}
	}
	
	/**
	 * Selects (checks) all of the buttons.
	 */
	public void checkAll() {
		for (int i = 0; i < buttons.size(); i++) {
			((AbstractButton) buttons.get(i)).setSelected(true);
		}
		repaint();
	}
	
	/**
	 * Selects (checks) none of the buttons.
	 */
	public void checkNone() {
		for (int i = 0; i < buttons.size(); i++) {
			((AbstractButton) buttons.get(i)).setSelected(false);
		}
		repaint();
	}
	
	public AbstractButton getButton(Object element) {
		AbstractButton button = null;
		for (int i = 0; i < getModel().getSize(); i++) {
			Object obj = getModel().getElementAt(i);
			if ((obj != null) && obj.equals(element)) {
				if (i < buttons.size()) {
					button = (AbstractButton) buttons.get(i);
				}
				break;
			}
		}
		return button;
	}
	
	public AbstractButton getButton(int index) {
		AbstractButton button = null;
		if ((index >= 0) && (index < buttons.size())) {
			button = (AbstractButton) buttons.get(index);
		}
		return button;
	}
	
	////////////////////////////////////////////
	// EVENT HANDLERS
	////////////////////////////////////////////
	
	public void itemStateChanged(ItemEvent e) {
		fireItemSelected(e);
	}
	
	public void valueChanged(ListSelectionEvent e) {
		if (!singleClickSelectionChange) {
			selectionChanging = e.getValueIsAdjusting();
		}
	}
	
	public void contentsChanged(ListDataEvent e) {
		if (e.getIndex1() >= e.getIndex0()) {
			for (int index = e.getIndex1(); index >= e.getIndex0(); index--) {
				if (index < buttons.size()) {
					Object obj = getModel().getElementAt(index);
					String text = (obj != null ? obj.toString() : "null");
					((AbstractButton) buttons.get(index)).setText(text);
				}
			}
			repaint(getCellBounds(e.getIndex0(), e.getIndex1()));
		}
	}
	
	public void intervalAdded(ListDataEvent e) {
		if (e.getIndex1() >= e.getIndex0()) {
			for (int index = e.getIndex1(); index >= e.getIndex0(); index--) {
				AbstractButton button = createButton(getModel().getElementAt(index));
				addButton(index, button);
			}
			repaint();
		}
	}
	public void intervalRemoved(ListDataEvent e) {
		if (e.getIndex1() >= e.getIndex0()) {
			for (int index = e.getIndex1(); index >= e.getIndex0(); index--) {
				removeButton(index);
			}
			repaint();
		}
	}
	
	public void mouseClicked(MouseEvent e) {
	}

	/**
	 * Toggles the checkbox if the clicked checkbox is the selected (highlighted) one
	 * or if the click was right on top of the checkbox.
	 */
	public void mousePressed(MouseEvent e) {
		boolean clickedOnCB = (e.getX() < 16);	// checkbox and radiobutton size
		if (singleClickSelectionChange || clickedOnCB || !selectionChanging) {
			int index = locationToIndex(e.getPoint());
			if (index != -1) {
				Rectangle bounds = getCellBounds(index, index);
				if ((bounds != null) && bounds.contains(e.getPoint())) {
					toggleCheckBox(index);
				}
			}
		}
	}

	public void mouseReleased(MouseEvent e) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void keyTyped(KeyEvent e) {
	}

	/**
	 * Toggles the selected checkbox(es) if the space or enter key is pressed.
	 */
	public void keyPressed(KeyEvent e) {
		int code = e.getKeyCode();
		if ((code == KeyEvent.VK_SPACE) || (code == KeyEvent.VK_ENTER)) {
			int[] indeces = getSelectedIndices();
			for (int i = 0; i < indeces.length; i++) {
				toggleCheckBox(indeces[i]);
			}
		}
	}

	public void keyReleased(KeyEvent e) {
	}
	
	/////////////////////////////////////////////////////////
	// PRIVATE METHODS
	/////////////////////////////////////////////////////////
	
	protected void addButton(AbstractButton button) {
		addButton(buttons.size(), button);
	}
	
	protected void addButton(int index, AbstractButton button) {
		if ((index >= buttons.size())) {
			buttons.add(button);
		} else {
			buttons.add(index, button);
		}
		
		// This is commented out because if fires an event every time a checkbox/button is toggled
		// regardless of whether it was a user event (mouse click) or done programmatically.
		// Also - the ItemEvent item variable is set to the AbstractButton, not the model element
		//button.addItemListener(this);
		
		if (radioButtonMode) {
			group.add(button);
		}
	}
	
	protected void removeButton(int index) {
		if ((index >= 0) && (index < buttons.size())) {
			AbstractButton button = (AbstractButton) buttons.remove(index);
			button.removeItemListener(this);
			if (group.containsButton(button)) {
				group.remove(button);
			}
		}
	}

	protected void removeButton(AbstractButton button) {
		if (button != null) {
			buttons.remove(button);
			button.removeItemListener(this);
			if (group.containsButton(button)) {
				group.remove(button);
			}
		}
	}

	
	protected void fireItemSelected(int index) {
		if (listeners.size() > 0) {
			AbstractButton button = (AbstractButton) buttons.get(index);
			Object obj = getModel().getElementAt(index);
			int state = (button.isSelected() ? ItemEvent.SELECTED : ItemEvent.DESELECTED);
			ItemEvent ie = new ItemEvent(button, ItemEvent.ITEM_STATE_CHANGED, obj, state);
			fireItemSelected(ie);
		}
	}

	protected void fireItemSelected(ItemEvent ie) {
		if (listeners.size() > 0) {
			ArrayList/*<ItemListener>*/ clone = new ArrayList(listeners);
			for (int i = 0; i < clone.size(); i++) {
				((ItemListener)clone.get(i)).itemStateChanged(ie);
			}
		}
	}
	
	/**
	 * Toggles the checkbox with the given index.
	 * @param index
	 */
	protected void toggleCheckBox(int index) {
		if ((index >= 0) && (index < getModel().getSize())) {
			AbstractButton button = (AbstractButton) buttons.get(index);
			if (button.isEnabled()) {
				button.setSelected(!button.isSelected());
				fireItemSelected(index);
				repaint();
			}
		}		
	}
	
	protected void reload(boolean repaint) {
		// remove the listeners and from the group
		while (buttons.size() > 0) {
			removeButton((AbstractButton) buttons.get(0));
		}
		
		group.clearButtons();
		int size = getModel().getSize();
		buttons = new ArrayList/*<AbstractButton>*/(size);
		for (int i = 0; i < size; i++) {
			Object obj = getModel().getElementAt(i);
			AbstractButton button = (obj instanceof AbstractButton ? (AbstractButton) obj : createButton(obj));
			addButton(i, button);
		}
		if (repaint) {
			repaint();
		}
	}

	private AbstractButton createButton(Object obj) {
		AbstractButton button;
		try {
			// make a new button/checkbox
			button = (AbstractButton) buttonClass.newInstance();
		} catch (Exception e) {
			button = new JCheckBox();
		}
		String text = (obj != null ? obj.toString() : "null");
		button.setText(text);
		return button;
	}
	
	/**
	 * Renders an {@link AbstractButton} object.  This will usually be a {@link JCheckBox}, 
	 * but it could also be a {@link JRadioButton}.  
	 */
	protected class CheckBoxCellRenderer implements ListCellRenderer {

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean selected, boolean hasFocus) {
			AbstractButton button;
			if (value instanceof AbstractButton) {
				button = (AbstractButton) value;
				buttons.set(index, button);
			} else {
				if (index < buttons.size()) {
					button = (AbstractButton) buttons.get(index);
				} else {
					button = createButton(value);
					addButton(index, button);
				}
			}
			
			button.setBackground(selected ? getSelectionBackground() : getBackground());
			button.setForeground(selected ? getSelectionForeground() : getForeground());
			button.setFocusPainted(false);
			button.setEnabled(isEnabled() && button.isEnabled());
			button.setFont(getFont());
			button.setBorderPainted(true);
			button.setBorder(selected ? UIManager.getBorder("List.focusCellHighlightBorder") : BORDER_NONE);
			return button;
		}
	}
	
	/**
	 * Extends {@link ButtonGroup} to not allow duplicate buttons in the group.
	 * @author Chris Callendar
	 * @date 7-Dec-06
	 */
	private class ExtendedButtonGroup extends ButtonGroup {
		
		public void add(AbstractButton b) {
			if (!containsButton(b)) {
				super.add(b);
			}
		}
		
		public void selectNone() {
			for (int i = 0; i < buttons.size(); i++) {
				AbstractButton button = (AbstractButton) buttons.get(i);
				if (button.isSelected()) {
					button.setSelected(false);
				}
			}
		}

		public void selectAll() {
			for (int i = 0; i < buttons.size(); i++) {
				AbstractButton button = (AbstractButton) buttons.get(i);
				if (!button.isSelected()) {
					button.setSelected(true);
				}
			}
		}
		
		public boolean containsButton(AbstractButton button) {
			return buttons.contains(button);
		}
		
		public void clearButtons() {
			Vector clone = (Vector) buttons.clone();
			for (int i = 0; i < clone.size(); i++) {
				remove((AbstractButton) clone.get(i));
			}
		}
	}
	
	/** Creates a panel containing a sample JCheckBoxList */
	public static JPanel createDemoPanel() {
		JPanel pnl = new JPanel(new BorderLayout());
		JButtonList list = new JButtonList(createDemoModel());
		
		list.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				String checked = (e.getStateChange() == ItemEvent.SELECTED ? " CHECKED" : " UNCHECKED");
				AbstractButton btn = (AbstractButton) e.getSource();
				System.out.println("Item checked: " + btn.getText() + checked);
			}
		});
		
		pnl.add(createOptionsPanel(list), BorderLayout.NORTH);
		pnl.add(new JScrollPane(list), BorderLayout.CENTER);
		pnl.setBorder(BorderFactory.createTitledBorder("JCheckBoxList"));
		return pnl;
	}
	

	/** Creates a panel containing a sample JCheckBoxList */
	public static JPanel createSimpleDemoPanel() {
		JPanel pnl = new JPanel(new BorderLayout());
		DefaultListModel model = new DefaultListModel();
		for (int i = 0; i < 5; i++) {
			model.addElement("Button #" + i);
		}
		JButtonList list = new JButtonList(model);
		pnl.add(new JScrollPane(list), BorderLayout.CENTER);
		pnl.setBorder(BorderFactory.createTitledBorder("JCheckBoxList"));
		return pnl;
	}
	
	private static DefaultListModel createDemoModel() {
		DefaultListModel model = new DefaultListModel();
		for (int i = 0; i < 3; i++) {
			model.addElement("Button #" + i);
		}
		model.addElement(new JCheckBox("A JCheckBox  "));
		model.addElement(new JRadioButton("A JRadioButton  "));
		model.addElement(new JToggleButton("A JToggleButton  "));
		model.addElement(new JCheckBoxMenuItem("A JCheckBoxMenuItem"));
		return model;
	}

	private static JPanel createOptionsPanel(final JButtonList list) {
		JPanel pnl = new JPanel(new GridLayout(4, 1, 0, 2));
		JCheckBox chk = new JCheckBox("Single Click Selections");
		chk.setSelected(true);
		chk.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				list.setSingleClickSelectionChange((e.getStateChange() == ItemEvent.SELECTED));
			}
		});
		pnl.add(chk);
		chk = new JCheckBox("RadioButton mode (single value selection)");
		chk.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				list.setRadioButtonMode((e.getStateChange() == ItemEvent.SELECTED));
			}
		});
		pnl.add(chk);
		final JComboBox combo = new JComboBox(new Class[] { JCheckBox.class, JRadioButton.class, JCheckBoxMenuItem.class, JToggleButton.class });
		combo.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED) {
					list.setButtonClass((Class) combo.getSelectedItem());
				}
			}
		});
		pnl.add(combo);
		return pnl;
	}
	
	public static void main(String[] args) {
		JFrame frame = new JFrame("JCheckBoxList Demo");
		frame.setSize(250, 400);
		frame.setLocation(400, 200);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.getContentPane().add(createDemoPanel(), BorderLayout.CENTER);
		frame.pack();
		frame.setVisible(true);
	}

}
